@@ -46,7 +46,8 @@ def pytest_addoption(parser):
             default=False, dest="raise_operr",
             help="Show the interp-level OperationError in app-level tests")
 
-def pytest_funcarg__space(request):
+@pytest.fixture(scope='function')
+def space(request):
     from pypy.tool.pytest.objspace import gettestobjspace
     spaceconfig = getattr(request.cls, 'spaceconfig', {})
     return gettestobjspace(**spaceconfig)
@@ -158,7 +159,8 @@ class LazyObjSpaceGetter(object):
         return space
 
 
-def pytest_runtest_setup(__multicall__, item):
+@pytest.hookimpl(tryfirst=True)
+def pytest_runtest_setup(item):
     if isinstance(item, py.test.collect.Function):
         appclass = item.getparent(py.test.Class)
         if appclass is not None:
@@ -171,8 +173,6 @@ def pytest_runtest_setup(__multicall__, item):
                 appclass.obj.space = LazyObjSpaceGetter()
             appclass.obj.runappdirect = option.runappdirect
 
-    __multicall__.execute()
-
 
 def pytest_ignore_collect(path):
     return path.check(link=1)
