@@ -7,6 +7,7 @@ Generate Python bytecode from a Abstract Syntax Tree.
 # you figure out a way to remove them, great, but try a translation first,
 # please.
 
+from rpython.rlib.objectmodel import specialize
 from pypy.interpreter.astcompiler import ast, assemble, symtable, consts, misc
 from pypy.interpreter.astcompiler import optimize # For side effects
 from pypy.interpreter.pyparser.error import SyntaxError
@@ -321,7 +322,7 @@ class PythonCodeGenerator(assemble.PythonCodeMaker):
     def visit_ClassDef(self, cls):
         self.update_position(cls.lineno, True)
         self.visit_sequence(cls.decorator_list)
-        self.load_const(self.space.wrap(cls.name))
+        self.load_const(self.space.newtext(cls.name))
         self.visit_sequence(cls.bases)
         bases_count = len(cls.bases) if cls.bases is not None else 0
         self.emit_op_arg(ops.BUILD_TUPLE, bases_count)
@@ -610,7 +611,7 @@ class PythonCodeGenerator(assemble.PythonCodeMaker):
                 level = 0
             else:
                 level = -1
-            self.load_const(self.space.wrap(level))
+            self.load_const(self.space.newint(level))
             self.load_const(self.space.w_None)
             self.emit_op_name(ops.IMPORT_NAME, self.names, alias.name)
             # If there's no asname then we store the root module.  If there is
@@ -653,12 +654,12 @@ class PythonCodeGenerator(assemble.PythonCodeMaker):
             level = -1
         else:
             level = imp.level
-        self.load_const(space.wrap(level))
+        self.load_const(space.newint(level))
         names_w = [None]*len(imp.names)
         for i in range(len(imp.names)):
             alias = imp.names[i]
             assert isinstance(alias, ast.alias)
-            names_w[i] = space.wrap(alias.name)
+            names_w[i] = space.newtext(alias.name)
         self.load_const(space.newtuple(names_w))
         if imp.module:
             mod_name = imp.module
@@ -943,7 +944,7 @@ class PythonCodeGenerator(assemble.PythonCodeMaker):
         self.name_op(name.id, name.ctx)
 
     def visit_keyword(self, keyword):
-        self.load_const(self.space.wrap(keyword.arg))
+        self.load_const(self.space.newtext(keyword.arg))
         keyword.value.walkabout(self)
 
     def visit_Call(self, call):
@@ -1027,11 +1028,17 @@ class PythonCodeGenerator(assemble.PythonCodeMaker):
 
     def visit_ListComp(self, lc):
         self.update_position(lc.lineno)
-        if len(lc.generators) != 1 or lc.generators[0].ifs:
+        if len(lc.generators) == 1:
+            comp = lc.generators[0]
+            assert isinstance(comp, ast.comprehension)
+            if comp.ifs:
+                single = False
+                self.emit_op_arg(ops.BUILD_LIST, 0)
+            else:
+                single = True
+        else:
             single = False
             self.emit_op_arg(ops.BUILD_LIST, 0)
-        else:
-            single = True
         self._listcomp_generator(lc.generators, 0, lc.elt, single=single)
 
     def _comp_generator(self, node, generators, gen_index):
