@@ -66,13 +66,13 @@ class W_CData(W_Root):
             from pypy.module._cffi_backend import ctypestruct
             if isinstance(self.ctype, ctypestruct.W_CTypeStructOrUnion):
                 extra1 = ' &'
-        return self.space.wrap("<cdata '%s%s' %s>" % (
+        return self.space.newtext("<cdata '%s%s' %s>" % (
             self.ctype.name, extra1, extra2))
 
     def nonzero(self):
         with self as ptr:
             nonzero = self.ctype.nonzero(ptr)
-        return self.space.wrap(nonzero)
+        return self.space.newbool(nonzero)
 
     def int(self, space):
         with self as ptr:
@@ -95,31 +95,43 @@ class W_CData(W_Root):
         from pypy.module._cffi_backend import ctypearray
         space = self.space
         if isinstance(self.ctype, ctypearray.W_CTypeArray):
-            return space.wrap(self.get_array_length())
+            return space.newint(self.get_array_length())
         raise oefmt(space.w_TypeError,
                     "cdata of type '%s' has no len()", self.ctype.name)
 
+    def _compare_mode(self, w_other):
+        from pypy.module._cffi_backend.ctypeprim import W_CTypePrimitive
+        zero = rffi.cast(lltype.Unsigned, 0)
+        self_is_ptr = not isinstance(self.ctype, W_CTypePrimitive)
+        other_is_ptr = (isinstance(w_other, W_CData) and
+                           not isinstance(w_other.ctype, W_CTypePrimitive))
+        if other_is_ptr and self_is_ptr:
+            with self as ptr1, w_other as ptr2:
+                ptr1 = rffi.cast(lltype.Unsigned, ptr1)
+                ptr2 = rffi.cast(lltype.Unsigned, ptr2)
+            return (0, ptr1, ptr2, None, None)
+        elif other_is_ptr or self_is_ptr:
+            return (-1, zero, zero, None, None)
+        else:
+            w_ob1 = self.convert_to_object()
+            if isinstance(w_other, W_CData):
+                w_ob2 = w_other.convert_to_object()
+            else:
+                w_ob2 = w_other
+            return (1, zero, zero, w_ob1, w_ob2)
+
     def _make_comparison(name):
         op = getattr(operator, name)
-        requires_ordering = name not in ('eq', 'ne')
         #
         def _cmp(self, w_other):
-            from pypy.module._cffi_backend.ctypeprim import W_CTypePrimitive
             space = self.space
-            if not isinstance(w_other, W_CData):
+            mode, adr1, adr2, w_ob1, w_ob2 = self._compare_mode(w_other)
+            if mode == 0:
+                return space.newbool(op(adr1, adr2))
+            elif mode == 1:
+                return getattr(space, name)(w_ob1, w_ob2)
+            else:
                 return space.w_NotImplemented
-
-            with self as ptr1, w_other as ptr2:
-                if requires_ordering:
-                    if (isinstance(self.ctype, W_CTypePrimitive) or
-                        isinstance(w_other.ctype, W_CTypePrimitive)):
-                        raise oefmt(space.w_TypeError,
-                                    "cannot do comparison on a primitive "
-                                    "cdata")
-                    ptr1 = rffi.cast(lltype.Unsigned, ptr1)
-                    ptr2 = rffi.cast(lltype.Unsigned, ptr2)
-                result = op(ptr1, ptr2)
-            return space.newbool(result)
         #
         return func_with_new_name(_cmp, name)
 
@@ -131,13 +143,18 @@ class W_CData(W_Root):
     ge = _make_comparison('ge')
 
     def hash(self):
+        from pypy.module._cffi_backend.ctypeprim import W_CTypePrimitive
+        if isinstance(self.ctype, W_CTypePrimitive):
+            w_ob = self.convert_to_object()
+            if not isinstance(w_ob, W_CData):
+                return self.space.hash(w_ob)
         ptr = self.unsafe_escaping_ptr()
         h = rffi.cast(lltype.Signed, ptr)
         # To hash pointers in dictionaries.  Assumes that h shows some
         # alignment (to 4, 8, maybe 16 bytes), so we use the following
         # formula to avoid the trailing bits being always 0.
         h = h ^ (h >> 4)
-        return self.space.wrap(h)
+        return self.space.newint(h)
 
     def getitem(self, w_index):
         space = self.space
@@ -232,10 +249,10 @@ class W_CData(W_Root):
         from pypy.module._cffi_backend import ctypeprim
         space = self.space
         if isinstance(ctitem, ctypeprim.W_CTypePrimitive) and ctitem.size == 1:
-            if space.isinstance_w(w_value, space.w_str):
+            if space.isinstance_w(w_value, space.w_bytes):
                 from rpython.rtyper.annlowlevel import llstr
                 from rpython.rtyper.lltypesystem.rstr import copy_string_to_raw
-                value = space.str_w(w_value)
+                value = space.bytes_w(w_value)
                 if len(value) != length:
                     raise oefmt(space.w_ValueError,
                                 "need a string of length %d, got %d",
@@ -319,21 +336,32 @@ class W_CData(W_Root):
                         "pointer subtraction: the distance between the two "
                         "pointers is not a multiple of the item size")
                 diff //= itemsize
-            return space.wrap(diff)
+            return space.newint(diff)
         #
         return self._add_or_sub(w_other, -1)
 
-    def getcfield(self, w_attr):
-        return self.ctype.getcfield(self.space.str_w(w_attr))
+    def getcfield(self, w_attr, mode):
+        space = self.space
+        attr = space.text_w(w_attr)
+        try:
+            cfield = self.ctype.getcfield(attr)
+        except KeyError:
+            raise oefmt(space.w_AttributeError, "cdata '%s' has no field '%s'",
+                        self.ctype.name, attr)
+        if cfield is None:
+            raise oefmt(space.w_AttributeError,
+                        "cdata '%s' points to an opaque type: cannot %s fields",
+                        self.ctype.name, mode)
+        return cfield
 
     def getattr(self, w_attr):
-        cfield = self.getcfield(w_attr)
+        cfield = self.getcfield(w_attr, mode="read")
         with self as ptr:
-            w_res = cfield.read(ptr)
+            w_res = cfield.read(ptr, self)
         return w_res
 
     def setattr(self, w_attr, w_value):
-        cfield = self.getcfield(w_attr)
+        cfield = self.getcfield(w_attr, mode="write")
         with self as ptr:
             cfield.write(ptr, w_value)
 
@@ -397,7 +425,7 @@ class W_CData(W_Root):
         space = self.space
         if space.is_none(w_destructor):
             if isinstance(self, W_CDataGCP):
-                self.w_destructor = None
+                self.detach_destructor()
                 return space.w_None
             raise oefmt(space.w_TypeError,
                         "Can remove destructor only on a object "
@@ -419,8 +447,8 @@ class W_CData(W_Root):
         with self as ptr:
             if not ptr:
                 raise oefmt(space.w_RuntimeError,
-                            "cannot use unpack() on %s",
-                            space.str_w(self.repr()))
+                            "cannot use unpack() on %R",
+                            self)
             w_result = ctype.ctitem.unpack_ptr(ctype, ptr, length)
         return w_result
 
@@ -430,7 +458,10 @@ class W_CData(W_Root):
         if isinstance(ct, W_CTypePointer):
             ct = ct.ctitem
         lst = ct.cdata_dir()
-        return space.newlist([space.wrap(s) for s in lst])
+        return space.newlist([space.newtext(s) for s in lst])
+
+    def get_structobj(self):
+        return None
 
 
 class W_CDataMem(W_CData):
@@ -453,28 +484,36 @@ class W_CDataNewOwning(W_CData):
     by newp().  They create and free their own memory according to an
     allocator."""
 
-    # the 'length' is either >= 0 for arrays, or -1 for pointers.
-    _attrs_ = ['length']
-    _immutable_fields_ = ['length']
+    # the 'allocated_length' is >= 0 for arrays; for var-sized
+    # structures it is the total size in bytes; otherwise it is -1.
+    _attrs_ = ['allocated_length']
+    _immutable_fields_ = ['allocated_length']
 
     def __init__(self, space, cdata, ctype, length=-1):
         W_CData.__init__(self, space, cdata, ctype)
-        self.length = length
+        self.allocated_length = length
 
     def _repr_extra(self):
         return self._repr_extra_owning()
 
     def _sizeof(self):
         ctype = self.ctype
-        if self.length >= 0:
+        if self.allocated_length >= 0:
             from pypy.module._cffi_backend import ctypearray
-            assert isinstance(ctype, ctypearray.W_CTypeArray)
-            return self.length * ctype.ctitem.size
+            if isinstance(ctype, ctypearray.W_CTypeArray):
+                return self.allocated_length * ctype.ctitem.size
+            else:
+                return self.allocated_length    # var-sized struct size
         else:
             return ctype.size
 
     def get_array_length(self):
-        return self.length
+        from pypy.module._cffi_backend import ctypearray
+        assert isinstance(self.ctype, ctypearray.W_CTypeArray)
+        return self.allocated_length
+
+    def get_structobj(self):
+        return self
 
 
 class W_CDataNewStd(W_CDataNewOwning):
@@ -508,12 +547,19 @@ class W_CDataPtrToStructOrUnion(W_CData):
         self.structobj = structobj
 
     def _repr_extra(self):
-        return self._repr_extra_owning()
+        return self.structobj._repr_extra_owning()
 
     def _do_getitem(self, ctype, i):
         assert i == 0
         return self.structobj
 
+    def get_structobj(self):
+        structobj = self.structobj
+        if isinstance(structobj, W_CDataNewOwning):
+            return structobj
+        else:
+            return None
+
 
 class W_CDataSliced(W_CData):
     """Subclass with an explicit length, for slices."""
@@ -547,7 +593,7 @@ class W_CDataHandle(W_CData):
 
     def _repr_extra(self):
         w_repr = self.space.repr(self.w_keepalive)
-        return "handle to %s" % (self.space.str_w(w_repr),)
+        return "handle to %s" % (self.space.text_w(w_repr),)
 
 
 class W_CDataFromBuffer(W_CData):
@@ -586,6 +632,10 @@ class W_CDataGCP(W_CData):
             self.w_destructor = None
             self.space.call_function(w_destructor, self.w_original_cdata)
 
+    def detach_destructor(self):
+        self.w_destructor = None
+        self.may_unregister_rpython_finalizer(self.space)
+
 
 W_CData.typedef = TypeDef(
     '_cffi_backend.CData',
