@@ -42,16 +42,16 @@ class CodecState(object):
                 w_input = space.newunicode(input)
             w_exc =  space.call_function(
                 w_cls,
-                space.wrap(encoding),
+                space.newtext(encoding),
                 w_input,
-                space.wrap(startpos),
-                space.wrap(endpos),
-                space.wrap(reason))
+                space.newint(startpos),
+                space.newint(endpos),
+                space.newtext(reason))
             w_res = space.call_function(w_errorhandler, w_exc)
             if (not space.isinstance_w(w_res, space.w_tuple)
                 or space.len_w(w_res) != 2
                 or not space.isinstance_w(
-                                 space.getitem(w_res, space.wrap(0)),
+                                 space.getitem(w_res, space.newint(0)),
                                  space.w_unicode)):
                 raise oefmt(space.w_TypeError,
                             "%s error handler must return (unicode, int) "
@@ -86,7 +86,7 @@ class CodecState(object):
             return self.unicodedata_handler
         try:
             w_unicodedata = space.getbuiltinmodule("unicodedata")
-            w_getcode = space.getattr(w_unicodedata, space.wrap("_get_code"))
+            w_getcode = space.getattr(w_unicodedata, space.newtext("_get_code"))
         except OperationError:
             return None
         else:
@@ -121,7 +121,7 @@ def register_codec(space, w_search_function):
         raise oefmt(space.w_TypeError, "argument must be callable")
 
 
-@unwrap_spec(encoding=str)
+@unwrap_spec(encoding='text')
 def lookup_codec(space, encoding):
     """lookup(encoding) -> (encoder, decoder, stream_reader, stream_writer)
     Looks up a codec tuple in the Python codec registry and returns
@@ -140,9 +140,9 @@ def lookup_codec(space, encoding):
 def _lookup_codec_loop(space, encoding, normalized_encoding):
     state = space.fromcache(CodecState)
     if state.codec_need_encodings:
-        w_import = space.getattr(space.builtin, space.wrap("__import__"))
+        w_import = space.getattr(space.builtin, space.newtext("__import__"))
         # registers new codecs
-        space.call_function(w_import, space.wrap("encodings"))
+        space.call_function(w_import, space.newtext("encodings"))
         state.codec_need_encodings = False
         if len(state.codec_search_path) == 0:
             raise oefmt(space.w_LookupError,
@@ -150,7 +150,7 @@ def _lookup_codec_loop(space, encoding, normalized_encoding):
                         "encoding")
     for w_search in state.codec_search_path:
         w_result = space.call_function(w_search,
-                                       space.wrap(normalized_encoding))
+                                       space.newtext(normalized_encoding))
         if not space.is_w(w_result, space.w_None):
             if not (space.isinstance_w(w_result, space.w_tuple) and
                     space.len_w(w_result) == 4):
@@ -167,16 +167,16 @@ def _lookup_codec_loop(space, encoding, normalized_encoding):
 
 def check_exception(space, w_exc):
     try:
-        w_start = space.getattr(w_exc, space.wrap('start'))
-        w_end = space.getattr(w_exc, space.wrap('end'))
-        w_obj = space.getattr(w_exc, space.wrap('object'))
+        w_start = space.getattr(w_exc, space.newtext('start'))
+        w_end = space.getattr(w_exc, space.newtext('end'))
+        w_obj = space.getattr(w_exc, space.newtext('object'))
     except OperationError as e:
         if not e.match(space, space.w_AttributeError):
             raise
         raise oefmt(space.w_TypeError, "wrong exception")
 
     delta = space.int_w(w_end) - space.int_w(w_start)
-    if delta < 0 or not (space.isinstance_w(w_obj, space.w_str) or
+    if delta < 0 or not (space.isinstance_w(w_obj, space.w_bytes) or
                          space.isinstance_w(w_obj, space.w_unicode)):
         raise oefmt(space.w_TypeError, "wrong exception")
 
@@ -189,23 +189,23 @@ def strict_errors(space, w_exc):
 
 def ignore_errors(space, w_exc):
     check_exception(space, w_exc)
-    w_end = space.getattr(w_exc, space.wrap('end'))
-    return space.newtuple([space.wrap(u''), w_end])
+    w_end = space.getattr(w_exc, space.newtext('end'))
+    return space.newtuple([space.newunicode(u''), w_end])
 
 def replace_errors(space, w_exc):
     check_exception(space, w_exc)
-    w_start = space.getattr(w_exc, space.wrap('start'))
-    w_end = space.getattr(w_exc, space.wrap('end'))
+    w_start = space.getattr(w_exc, space.newtext('start'))
+    w_end = space.getattr(w_exc, space.newtext('end'))
     size = space.int_w(w_end) - space.int_w(w_start)
     if space.isinstance_w(w_exc, space.w_UnicodeEncodeError):
         text = u'?' * size
-        return space.newtuple([space.wrap(text), w_end])
+        return space.newtuple([space.newunicode(text), w_end])
     elif space.isinstance_w(w_exc, space.w_UnicodeDecodeError):
         text = u'\ufffd'
-        return space.newtuple([space.wrap(text), w_end])
+        return space.newtuple([space.newunicode(text), w_end])
     elif space.isinstance_w(w_exc, space.w_UnicodeTranslateError):
         text = u'\ufffd' * size
-        return space.newtuple([space.wrap(text), w_end])
+        return space.newtuple([space.newunicode(text), w_end])
     else:
         raise oefmt(space.w_TypeError,
                     "don't know how to handle %T in error callback", w_exc)
@@ -213,9 +213,9 @@ def replace_errors(space, w_exc):
 def xmlcharrefreplace_errors(space, w_exc):
     check_exception(space, w_exc)
     if space.isinstance_w(w_exc, space.w_UnicodeEncodeError):
-        obj = space.realunicode_w(space.getattr(w_exc, space.wrap('object')))
-        start = space.int_w(space.getattr(w_exc, space.wrap('start')))
-        w_end = space.getattr(w_exc, space.wrap('end'))
+        obj = space.realunicode_w(space.getattr(w_exc, space.newtext('object')))
+        start = space.int_w(space.getattr(w_exc, space.newtext('start')))
+        w_end = space.getattr(w_exc, space.newtext('end'))
         end = space.int_w(w_end)
         builder = UnicodeBuilder()
         pos = start
@@ -231,7 +231,7 @@ def xmlcharrefreplace_errors(space, w_exc):
             builder.append(unicode(str(code)))
             builder.append(u";")
             pos += 1
-        return space.newtuple([space.wrap(builder.build()), w_end])
+        return space.newtuple([space.newunicode(builder.build()), w_end])
     else:
         raise oefmt(space.w_TypeError,
                     "don't know how to handle %T in error callback", w_exc)
@@ -239,9 +239,9 @@ def xmlcharrefreplace_errors(space, w_exc):
 def backslashreplace_errors(space, w_exc):
     check_exception(space, w_exc)
     if space.isinstance_w(w_exc, space.w_UnicodeEncodeError):
-        obj = space.realunicode_w(space.getattr(w_exc, space.wrap('object')))
-        start = space.int_w(space.getattr(w_exc, space.wrap('start')))
-        w_end = space.getattr(w_exc, space.wrap('end'))
+        obj = space.realunicode_w(space.getattr(w_exc, space.newtext('object')))
+        start = space.int_w(space.getattr(w_exc, space.newtext('start')))
+        w_end = space.getattr(w_exc, space.newtext('end'))
         end = space.int_w(w_end)
         builder = UnicodeBuilder()
         pos = start
@@ -263,7 +263,7 @@ def backslashreplace_errors(space, w_exc):
                 builder.append_multiple_char(u'0', nb)
             builder.append_slice(unicode(num), 2, lnum)
             pos += 1
-        return space.newtuple([space.wrap(builder.build()), w_end])
+        return space.newtuple([space.newunicode(builder.build()), w_end])
     else:
         raise oefmt(space.w_TypeError,
                     "don't know how to handle %T in error callback", w_exc)
@@ -274,10 +274,11 @@ def register_builtin_error_handlers(space):
     for error in ("strict", "ignore", "replace", "xmlcharrefreplace",
                   "backslashreplace"):
         name = error + "_errors"
-        state.codec_error_registry[error] = space.wrap(interp2app(globals()[name]))
+        state.codec_error_registry[error] = interp2app(
+                globals()[name]).spacebind(space)
 
 
-@unwrap_spec(errors=str)
+@unwrap_spec(errors='text')
 def lookup_error(space, errors):
     """lookup_error(errors) -> handler
 
@@ -294,7 +295,7 @@ def lookup_error(space, errors):
     return w_err_handler
 
 
-@unwrap_spec(errors=str)
+@unwrap_spec(errors='text')
 def encode(space, w_obj, w_encoding=None, errors='strict'):
     """encode(obj, [encoding[,errors]]) -> object
 
@@ -308,22 +309,22 @@ def encode(space, w_obj, w_encoding=None, errors='strict'):
     if w_encoding is None:
         encoding = space.sys.defaultencoding
     else:
-        encoding = space.str_w(w_encoding)
-    w_encoder = space.getitem(lookup_codec(space, encoding), space.wrap(0))
-    w_res = space.call_function(w_encoder, w_obj, space.wrap(errors))
-    return space.getitem(w_res, space.wrap(0))
+        encoding = space.text_w(w_encoding)
+    w_encoder = space.getitem(lookup_codec(space, encoding), space.newint(0))
+    w_res = space.call_function(w_encoder, w_obj, space.newtext(errors))
+    return space.getitem(w_res, space.newint(0))
 
-@unwrap_spec(errors='str_or_None')
+@unwrap_spec(errors='text_or_none')
 def readbuffer_encode(space, w_data, errors='strict'):
     s = space.getarg_w('s#', w_data)
-    return space.newtuple([space.newbytes(s), space.wrap(len(s))])
+    return space.newtuple([space.newbytes(s), space.newint(len(s))])
 
-@unwrap_spec(errors='str_or_None')
+@unwrap_spec(errors='text_or_none')
 def charbuffer_encode(space, w_data, errors='strict'):
     s = space.getarg_w('t#', w_data)
-    return space.newtuple([space.wrap(s), space.wrap(len(s))])
+    return space.newtuple([space.newbytes(s), space.newint(len(s))])
 
-@unwrap_spec(errors=str)
+@unwrap_spec(errors='text')
 def decode(space, w_obj, w_encoding=None, errors='strict'):
     """decode(obj, [encoding[,errors]]) -> object
 
@@ -337,18 +338,18 @@ def decode(space, w_obj, w_encoding=None, errors='strict'):
     if w_encoding is None:
         encoding = space.sys.defaultencoding
     else:
-        encoding = space.str_w(w_encoding)
-    w_decoder = space.getitem(lookup_codec(space, encoding), space.wrap(1))
+        encoding = space.text_w(w_encoding)
+    w_decoder = space.getitem(lookup_codec(space, encoding), space.newint(1))
     if space.is_true(w_decoder):
-        w_res = space.call_function(w_decoder, w_obj, space.wrap(errors))
+        w_res = space.call_function(w_decoder, w_obj, space.newtext(errors))
         if (not space.isinstance_w(w_res, space.w_tuple) or space.len_w(w_res) != 2):
             raise oefmt(space.w_TypeError,
                         "encoder must return a tuple (object, integer)")
-        return space.getitem(w_res, space.wrap(0))
+        return space.getitem(w_res, space.newint(0))
     else:
         assert 0, "XXX, what to do here?"
 
-@unwrap_spec(errors=str)
+@unwrap_spec(errors='text')
 def register_error(space, errors, w_handler):
     """register_error(errors, handler)
 
@@ -372,21 +373,21 @@ from rpython.rlib import runicode
 def make_encoder_wrapper(name):
     rname = "unicode_encode_%s" % (name.replace("_encode", ""), )
     assert hasattr(runicode, rname)
-    @unwrap_spec(uni=unicode, errors='str_or_None')
+    @unwrap_spec(uni=unicode, errors='text_or_none')
     def wrap_encoder(space, uni, errors="strict"):
         if errors is None:
             errors = 'strict'
         state = space.fromcache(CodecState)
         func = getattr(runicode, rname)
         result = func(uni, len(uni), errors, state.encode_error_handler)
-        return space.newtuple([space.newbytes(result), space.wrap(len(uni))])
+        return space.newtuple([space.newbytes(result), space.newint(len(uni))])
     wrap_encoder.func_name = rname
     globals()[name] = wrap_encoder
 
 def make_decoder_wrapper(name):
     rname = "str_decode_%s" % (name.replace("_decode", ""), )
     assert hasattr(runicode, rname)
-    @unwrap_spec(string='bufferstr', errors='str_or_None',
+    @unwrap_spec(string='bufferstr', errors='text_or_none',
                  w_final=WrappedDefault(False))
     def wrap_decoder(space, string, errors="strict", w_final=None):
         if errors is None:
@@ -396,7 +397,7 @@ def make_decoder_wrapper(name):
         func = getattr(runicode, rname)
         result, consumed = func(string, len(string), errors,
                                 final, state.decode_error_handler)
-        return space.newtuple([space.wrap(result), space.wrap(consumed)])
+        return space.newtuple([space.newunicode(result), space.newint(consumed)])
     wrap_decoder.func_name = rname
     globals()[name] = wrap_decoder
 
@@ -436,30 +437,36 @@ if hasattr(runicode, 'str_decode_mbcs'):
 
 # utf-8 functions are not regular, because we have to pass
 # "allow_surrogates=True"
-@unwrap_spec(uni=unicode, errors='str_or_None')
+@unwrap_spec(uni=unicode, errors='text_or_none')
 def utf_8_encode(space, uni, errors="strict"):
     if errors is None:
         errors = 'strict'
     state = space.fromcache(CodecState)
-    result = runicode.unicode_encode_utf_8(
+    # NB. can't call unicode_encode_utf_8() directly because that's
+    # an @elidable function nowadays.  Instead, we need the _impl().
+    # (The problem is the errorhandler, which calls arbitrary Python.)
+    result = runicode.unicode_encode_utf_8_impl(
         uni, len(uni), errors, state.encode_error_handler,
         allow_surrogates=True)
-    return space.newtuple([space.wrap(result), space.wrap(len(uni))])
+    return space.newtuple([space.newbytes(result), space.newint(len(uni))])
 
-@unwrap_spec(string='bufferstr', errors='str_or_None',
+@unwrap_spec(string='bufferstr', errors='text_or_none',
              w_final = WrappedDefault(False))
 def utf_8_decode(space, string, errors="strict", w_final=None):
     if errors is None:
         errors = 'strict'
     final = space.is_true(w_final)
     state = space.fromcache(CodecState)
-    result, consumed = runicode.str_decode_utf_8(
+    # NB. can't call str_decode_utf_8() directly because that's
+    # an @elidable function nowadays.  Instead, we need the _impl().
+    # (The problem is the errorhandler, which calls arbitrary Python.)
+    result, consumed = runicode.str_decode_utf_8_impl(
         string, len(string), errors,
         final, state.decode_error_handler,
         allow_surrogates=True)
-    return space.newtuple([space.wrap(result), space.wrap(consumed)])
+    return space.newtuple([space.newunicode(result), space.newint(consumed)])
 
-@unwrap_spec(data='bufferstr', errors='str_or_None', byteorder=int,
+@unwrap_spec(data='bufferstr', errors='text_or_none', byteorder=int,
              w_final=WrappedDefault(False))
 def utf_16_ex_decode(space, data, errors='strict', byteorder=0, w_final=None):
     if errors is None:
@@ -477,10 +484,10 @@ def utf_16_ex_decode(space, data, errors='strict', byteorder=0, w_final=None):
         consumed = 0
     res, consumed, byteorder = runicode.str_decode_utf_16_helper(
         data, len(data), errors, final, state.decode_error_handler, byteorder)
-    return space.newtuple([space.wrap(res), space.wrap(consumed),
-                           space.wrap(byteorder)])
+    return space.newtuple([space.newunicode(res), space.newint(consumed),
+                           space.newint(byteorder)])
 
-@unwrap_spec(data='bufferstr', errors='str_or_None', byteorder=int,
+@unwrap_spec(data='bufferstr', errors='text_or_none', byteorder=int,
              w_final=WrappedDefault(False))
 def utf_32_ex_decode(space, data, errors='strict', byteorder=0, w_final=None):
     final = space.is_true(w_final)
@@ -496,8 +503,8 @@ def utf_32_ex_decode(space, data, errors='strict', byteorder=0, w_final=None):
         consumed = 0
     res, consumed, byteorder = runicode.str_decode_utf_32_helper(
         data, len(data), errors, final, state.decode_error_handler, byteorder)
-    return space.newtuple([space.wrap(res), space.wrap(consumed),
-                           space.wrap(byteorder)])
+    return space.newtuple([space.newunicode(res), space.newint(consumed),
+                           space.newint(byteorder)])
 
 # ____________________________________________________________
 # Charmap
@@ -560,7 +567,7 @@ class Charmap_Encode:
                 raise
             return errorchar
 
-        if space.isinstance_w(w_ch, space.w_str):
+        if space.isinstance_w(w_ch, space.w_bytes):
             # Charmap may return a string
             return space.bytes_w(w_ch)
         elif space.isinstance_w(w_ch, space.w_int):
@@ -578,12 +585,12 @@ class Charmap_Encode:
             "character mapping must return integer, None or str")
 
 
-@unwrap_spec(string='bufferstr', errors='str_or_None')
+@unwrap_spec(string='bufferstr', errors='text_or_none')
 def charmap_decode(space, string, errors="strict", w_mapping=None):
     if errors is None:
         errors = 'strict'
     if len(string) == 0:
-        return space.newtuple([space.wrap(u''), space.wrap(0)])
+        return space.newtuple([space.newunicode(u''), space.newint(0)])
 
     if space.is_none(w_mapping):
         mapping = None
@@ -595,9 +602,9 @@ def charmap_decode(space, string, errors="strict", w_mapping=None):
     result, consumed = runicode.str_decode_charmap(
         string, len(string), errors,
         final, state.decode_error_handler, mapping)
-    return space.newtuple([space.wrap(result), space.wrap(consumed)])
+    return space.newtuple([space.newunicode(result), space.newint(consumed)])
 
-@unwrap_spec(uni=unicode, errors='str_or_None')
+@unwrap_spec(uni=unicode, errors='text_or_none')
 def charmap_encode(space, uni, errors="strict", w_mapping=None):
     if errors is None:
         errors = 'strict'
@@ -610,7 +617,7 @@ def charmap_encode(space, uni, errors="strict", w_mapping=None):
     result = runicode.unicode_encode_charmap(
         uni, len(uni), errors,
         state.encode_error_handler, mapping)
-    return space.newtuple([space.newbytes(result), space.wrap(len(uni))])
+    return space.newtuple([space.newbytes(result), space.newint(len(uni))])
 
 
 @unwrap_spec(chars=unicode)
@@ -633,14 +640,14 @@ class UnicodeData_Handler:
     def call(self, name):
         space = self.space
         try:
-            w_code = space.call_function(self.w_getcode, space.wrap(name))
+            w_code = space.call_function(self.w_getcode, space.newtext(name))
         except OperationError as e:
             if not e.match(space, space.w_KeyError):
                 raise
             return -1
         return space.int_w(w_code)
 
-@unwrap_spec(string='bufferstr', errors='str_or_None',
+@unwrap_spec(string='bufferstr', errors='text_or_none',
              w_final=WrappedDefault(False))
 def unicode_escape_decode(space, string, errors="strict", w_final=None):
     if errors is None:
@@ -655,12 +662,12 @@ def unicode_escape_decode(space, string, errors="strict", w_final=None):
         final, state.decode_error_handler,
         unicode_name_handler)
 
-    return space.newtuple([space.wrap(result), space.wrap(consumed)])
+    return space.newtuple([space.newunicode(result), space.newint(consumed)])
 
 # ____________________________________________________________
 # Unicode-internal
 
-@unwrap_spec(errors='str_or_None')
+@unwrap_spec(errors='text_or_none')
 def unicode_internal_decode(space, w_string, errors="strict"):
     if errors is None:
         errors = 'strict'
@@ -671,31 +678,31 @@ def unicode_internal_decode(space, w_string, errors="strict"):
     string = space.readbuf_w(w_string).as_str()
 
     if len(string) == 0:
-        return space.newtuple([space.wrap(u''), space.wrap(0)])
+        return space.newtuple([space.newunicode(u''), space.newint(0)])
 
     final = True
     state = space.fromcache(CodecState)
     result, consumed = runicode.str_decode_unicode_internal(
         string, len(string), errors,
         final, state.decode_error_handler)
-    return space.newtuple([space.wrap(result), space.wrap(consumed)])
+    return space.newtuple([space.newunicode(result), space.newint(consumed)])
 
 # ____________________________________________________________
 # support for the "string escape" codec
 # This is a bytes-to bytes transformation
 
-@unwrap_spec(data=str, errors='str_or_None')
+@unwrap_spec(data='bytes', errors='text_or_none')
 def escape_encode(space, data, errors='strict'):
     from pypy.objspace.std.bytesobject import string_escape_encode
     result = string_escape_encode(data, quote="'")
     start = 1
     end = len(result) - 1
     assert end >= 0
-    w_result = space.wrap(result[start:end])
-    return space.newtuple([w_result, space.wrap(len(data))])
+    w_result = space.newbytes(result[start:end])
+    return space.newtuple([w_result, space.newint(len(data))])
 
-@unwrap_spec(data=str, errors='str_or_None')
+@unwrap_spec(data='bytes', errors='text_or_none')
 def escape_decode(space, data, errors='strict'):
     from pypy.interpreter.pyparser.parsestring import PyString_DecodeEscape
     result = PyString_DecodeEscape(space, data, errors, None)
-    return space.newtuple([space.newbytes(result), space.wrap(len(data))])
+    return space.newtuple([space.newbytes(result), space.newint(len(data))])
