@@ -68,7 +68,7 @@ class Box(object):
         scalar = multiarray.get("scalar")
 
         ret = space.newtuple([scalar, space.newtuple(
-            [space.wrap(self._get_dtype(space)), space.wrap(self.raw_str())])])
+            [self._get_dtype(space), space.newtext(self.raw_str())])])
         return ret
 
 
@@ -196,11 +196,11 @@ class W_GenericBox(W_NumpyObject):
 
     def descr_str(self, space):
         tp = self.get_dtype(space).itemtype
-        return space.wrap(tp.str_format(self, add_quotes=False))
+        return space.newtext(tp.str_format(self, add_quotes=False))
 
     def descr_repr(self, space):
         tp = self.get_dtype(space).itemtype
-        return space.wrap(tp.str_format(self, add_quotes=True))
+        return space.newtext(tp.str_format(self, add_quotes=True))
 
     def descr_format(self, space, w_spec):
         return space.format(self.item(space), w_spec)
@@ -209,7 +209,7 @@ class W_GenericBox(W_NumpyObject):
         return space.hash(self.item(space))
 
     def descr___array_priority__(self, space):
-        return space.wrap(0.0)
+        return space.newfloat(0.0)
 
     def descr_index(self, space):
         return space.index(self.item(space))
@@ -242,7 +242,7 @@ class W_GenericBox(W_NumpyObject):
         return space.hex(self.descr_int(space))
 
     def descr_nonzero(self, space):
-        return space.wrap(self.get_dtype(space).itemtype.bool(self))
+        return space.newbool(self.get_dtype(space).itemtype.bool(self))
 
     # TODO: support all kwargs in ufuncs like numpy ufunc_object.c
     sig = None
@@ -385,16 +385,16 @@ class W_GenericBox(W_NumpyObject):
         return self.get_dtype(space)
 
     def descr_get_size(self, space):
-        return space.wrap(1)
+        return space.newint(1)
 
     def descr_get_itemsize(self, space):
-        return space.wrap(self.get_dtype(space).elsize)
+        return space.newint(self.get_dtype(space).elsize)
 
     def descr_get_shape(self, space):
         return space.newtuple([])
 
     def descr_get_ndim(self, space):
-        return space.wrap(0)
+        return space.newint(0)
 
     def descr_copy(self, space):
         return self.convert_to(space, self.get_dtype(space))
@@ -412,18 +412,18 @@ class W_GenericBox(W_NumpyObject):
         return self.get_dtype(space).itemtype.byteswap(self)
 
     def descr_tostring(self, space, __args__):
-        w_meth = space.getattr(self.descr_ravel(space), space.wrap('tostring'))
+        w_meth = space.getattr(self.descr_ravel(space), space.newtext('tostring'))
         return space.call_args(w_meth, __args__)
 
     def descr_reshape(self, space, __args__):
-        w_meth = space.getattr(self.descr_ravel(space), space.wrap('reshape'))
+        w_meth = space.getattr(self.descr_ravel(space), space.newtext('reshape'))
         w_res = space.call_args(w_meth, __args__)
         if isinstance(w_res, W_NDimArray) and len(w_res.get_shape()) == 0:
             return w_res.get_scalar_value()
         return w_res
 
     def descr_nd_nonzero(self, space, __args__):
-        w_meth = space.getattr(self.descr_ravel(space), space.wrap('nonzero'))
+        w_meth = space.getattr(self.descr_ravel(space), space.newtext('nonzero'))
         return space.call_args(w_meth, __args__)
 
     def descr_get_real(self, space):
@@ -556,7 +556,7 @@ class W_FlexibleBox(W_GenericBox):
 class W_VoidBox(W_FlexibleBox):
     def descr_getitem(self, space, w_item):
         if space.isinstance_w(w_item, space.w_basestring):
-            item = space.str_w(w_item)
+            item = space.text_w(w_item)
         elif space.isinstance_w(w_item, space.w_int):
             indx = space.int_w(w_item)
             try:
@@ -579,7 +579,7 @@ class W_VoidBox(W_FlexibleBox):
             read_val = dtype.read(self.arr, self.ofs, ofs)
         if isinstance (read_val, W_StringBox):
             # StringType returns a str
-            return space.wrap(dtype.itemtype.to_str(read_val))
+            return space.newbytes(dtype.itemtype.to_str(read_val))
         return read_val
 
     def descr_iter(self, space):
@@ -587,7 +587,7 @@ class W_VoidBox(W_FlexibleBox):
 
     def descr_setitem(self, space, w_item, w_value):
         if space.isinstance_w(w_item, space.w_basestring):
-            item = space.str_w(w_item)
+            item = space.text_w(w_item)
         elif space.isinstance_w(w_item, space.w_int):
             indx = space.int_w(w_item)
             try:
@@ -613,7 +613,8 @@ class W_VoidBox(W_FlexibleBox):
 
 class W_CharacterBox(W_FlexibleBox):
     def convert_to(self, space, dtype):
-        return dtype.coerce(space, space.wrap(self.raw_str()))
+        # XXX should be newbytes?
+        return dtype.coerce(space, space.newtext(self.raw_str()))
 
     def descr_len(self, space):
         return space.len(self.item(space))
@@ -621,7 +622,7 @@ class W_CharacterBox(W_FlexibleBox):
 class W_StringBox(W_CharacterBox):
     def descr__new__string_box(space, w_subtype, w_arg):
         from pypy.module.micronumpy.descriptor import new_string_dtype
-        arg = space.str_w(space.str(w_arg))
+        arg = space.text_w(space.str(w_arg))
         arr = VoidBoxStorage(len(arg), new_string_dtype(space, len(arg)))
         for i in range(len(arg)):
             arr.storage[i] = arg[i]
@@ -635,7 +636,7 @@ class W_UnicodeBox(W_CharacterBox):
         if dtype.is_unicode():
             return self
         elif dtype.is_object():
-            return W_ObjectBox(space.wrap(self._value))
+            return W_ObjectBox(space.newunicode(self._value))
         else:
             raise oefmt(space.w_NotImplementedError,
                         "Conversion from unicode not implemented yet")
