@@ -75,7 +75,7 @@ class W_AbstractLongObject(W_Root):
         """
         bigint = space.bigint_w(self)
         try:
-            return space.wrap(bigint.bit_length())
+            return space.newint(bigint.bit_length())
         except OverflowError:
             raise oefmt(space.w_OverflowError, "too many digits in integer")
 
@@ -120,24 +120,24 @@ class W_AbstractLongObject(W_Root):
     def descr_get_imag(self, space):
         return space.newlong(0)
 
-    def _make_descr_unaryop(opname):
+    def _make_descr_unaryop_text(opname):
         op = getattr(rbigint, opname)
         @func_renamer('descr_' + opname)
         def descr_unaryop(self, space):
-            return space.wrap(op(self.asbigint()))
+            return space.newtext(op(self.asbigint()))
         descr_unaryop.__doc__ = 'x.__%s__(y) <==> %s(x, y)' % (opname, opname)
         return descr_unaryop
 
-    descr_repr = _make_descr_unaryop('repr')
-    descr_str = _make_descr_unaryop('str')
+    descr_repr = _make_descr_unaryop_text('repr')
+    descr_str = _make_descr_unaryop_text('str')
 
     def descr_hash(self, space):
         h = self.asbigint().hash()
         h -= (h == -1)
         return space.newint(h)
 
-    descr_oct = _make_descr_unaryop('oct')
-    descr_hex = _make_descr_unaryop('hex')
+    descr_oct = _make_descr_unaryop_text('oct')
+    descr_hex = _make_descr_unaryop_text('hex')
 
     def descr_pow(self, space, w_exponent, w_modulus=None):
         """x.__pow__(y[, z]) <==> pow(x, y[, z])"""
@@ -564,18 +564,24 @@ def descr__new__(space, w_longtype, w_x, w_base=None):
         elif (space.lookup(w_value, '__long__') is not None or
               space.lookup(w_value, '__int__') is not None):
             w_obj = space.long(w_value)
+            if (space.is_w(w_longtype, space.w_long) and
+                space.isinstance_w(w_obj, space.w_long)):
+                return w_obj
             return newbigint(space, w_longtype, space.bigint_w(w_obj))
         elif space.lookup(w_value, '__trunc__') is not None:
             w_obj = space.trunc(w_value)
+            if (space.is_w(w_longtype, space.w_long) and
+                space.isinstance_w(w_obj, space.w_long)):
+                return w_obj
             # :-(  blame CPython 2.7
             if space.lookup(w_obj, '__long__') is not None:
                 w_obj = space.long(w_obj)
             else:
                 w_obj = space.int(w_obj)
             return newbigint(space, w_longtype, space.bigint_w(w_obj))
-        elif space.isinstance_w(w_value, space.w_str):
+        elif space.isinstance_w(w_value, space.w_bytes):
             return _string_to_w_long(space, w_longtype, w_value,
-                                     space.str_w(w_value))
+                                     space.bytes_w(w_value))
         elif space.isinstance_w(w_value, space.w_unicode):
             from pypy.objspace.std.unicodeobject import unicode_to_decimal_w
             return _string_to_w_long(space, w_longtype, w_value,
@@ -599,7 +605,7 @@ def descr__new__(space, w_longtype, w_x, w_base=None):
             s = unicode_to_decimal_w(space, w_value)
         else:
             try:
-                s = space.str_w(w_value)
+                s = space.bytes_w(w_value)
             except OperationError:
                 raise oefmt(space.w_TypeError,
                             "long() can't convert non-string with explicit "
