@@ -1,7 +1,8 @@
 """The builtin unicode implementation"""
 
 from rpython.rlib.objectmodel import (
-    compute_hash, compute_unique_id, import_from_mixin)
+    compute_hash, compute_unique_id, import_from_mixin,
+    enforceargs)
 from rpython.rlib.buffer import StringBuffer
 from rpython.rlib.rstring import StringBuilder, UnicodeBuilder
 from rpython.rlib.runicode import (
@@ -29,6 +30,7 @@ class W_UnicodeObject(W_Root):
     import_from_mixin(StringMethods)
     _immutable_fields_ = ['_value']
 
+    @enforceargs(uni=unicode)
     def __init__(self, unistr):
         assert isinstance(unistr, unicode)
         self._value = unistr
@@ -72,10 +74,10 @@ class W_UnicodeObject(W_Root):
             else:
                 base = 257       # empty unicode string: base value 257
             uid = (base << IDTAG_SHIFT) | IDTAG_SPECIAL
-        return space.wrap(uid)
+        return space.newint(uid)
 
     def str_w(self, space):
-        return space.str_w(space.str(self))
+        return space.text_w(space.str(self))
 
     def unicode_w(self, space):
         return self._value
@@ -101,7 +103,7 @@ class W_UnicodeObject(W_Root):
             raise oefmt(space.w_TypeError,
                          "ord() expected a character, but string of length %d "
                          "found", len(self._value))
-        return space.wrap(ord(self._value[0]))
+        return space.newint(ord(self._value[0]))
 
     def _new(self, value):
         return W_UnicodeObject(value)
@@ -124,11 +126,14 @@ class W_UnicodeObject(W_Root):
         return True
 
     @staticmethod
-    def _op_val(space, w_other):
+    def _op_val(space, w_other, strict=None):
         if isinstance(w_other, W_UnicodeObject):
             return w_other._value
-        if space.isinstance_w(w_other, space.w_str):
+        if space.isinstance_w(w_other, space.w_bytes):
             return unicode_from_string(space, w_other)._value
+        if strict:
+            raise oefmt(space.w_TypeError,
+                "%s arg must be None, unicode or str", strict)
         return unicode_from_encoded_object(
             space, w_other, None, "strict")._value
 
@@ -198,7 +203,7 @@ class W_UnicodeObject(W_Root):
         is_precisely_unicode = space.is_w(space.type(w_obj), space.w_unicode)
         if (is_precisely_unicode or
             (space.isinstance_w(w_obj, space.w_unicode) and
-             space.findattr(w_obj, space.wrap('__unicode__')) is None)):
+             space.findattr(w_obj, space.newtext('__unicode__')) is None)):
             if encoding is not None or errors is not None:
                 raise oefmt(space.w_TypeError,
                             "decoding Unicode is not supported")
@@ -224,14 +229,14 @@ class W_UnicodeObject(W_Root):
         chars = self._value
         size = len(chars)
         s = _repr_function(chars, size, "strict")
-        return space.wrap(s)
+        return space.newtext(s)
 
     def descr_str(self, space):
         return encode_object(space, self, None, None)
 
     def descr_hash(self, space):
         x = compute_hash(self._value)
-        return space.wrap(x)
+        return space.newint(x)
 
     def descr_eq(self, space, w_other):
         try:
@@ -244,7 +249,7 @@ class W_UnicodeObject(W_Root):
                 msg = ("Unicode equal comparison failed to convert both "
                        "arguments to Unicode - interpreting them as being "
                        "unequal")
-                space.warn(space.wrap(msg), space.w_UnicodeWarning)
+                space.warn(space.newtext(msg), space.w_UnicodeWarning)
                 return space.w_False
             raise
         return space.newbool(res)
@@ -260,7 +265,7 @@ class W_UnicodeObject(W_Root):
                 msg = ("Unicode unequal comparison failed to convert both "
                        "arguments to Unicode - interpreting them as being "
                        "unequal")
-                space.warn(space.wrap(msg), space.w_UnicodeWarning)
+                space.warn(space.newtext(msg), space.w_UnicodeWarning)
                 return space.w_True
             raise
         return space.newbool(res)
@@ -316,15 +321,18 @@ class W_UnicodeObject(W_Root):
     def descr_mod(self, space, w_values):
         return mod_format(space, self, w_values, do_unicode=True)
 
+    def descr_rmod(self, space, w_values):
+        return mod_format(space, w_values, self, do_unicode=True)
+
     def descr_translate(self, space, w_table):
         selfvalue = self._value
         w_sys = space.getbuiltinmodule('sys')
         maxunicode = space.int_w(space.getattr(w_sys,
-                                               space.wrap("maxunicode")))
+                                               space.newtext("maxunicode")))
         result = []
         for unichar in selfvalue:
             try:
-                w_newval = space.getitem(w_table, space.wrap(ord(unichar)))
+                w_newval = space.getitem(w_table, space.newint(ord(unichar)))
             except OperationError as e:
                 if e.match(space, space.w_LookupError):
                     result.append(unichar)
@@ -358,15 +366,15 @@ class W_UnicodeObject(W_Root):
         l = space.listview_unicode(w_list)
         if l is not None:
             if len(l) == 1:
-                return space.wrap(l[0])
-            return space.wrap(self._val(space).join(l))
+                return space.newunicode(l[0])
+            return space.newunicode(self._val(space).join(l))
         return self._StringMethods_descr_join(space, w_list)
 
     def _join_return_one(self, space, w_obj):
         return space.is_w(space.type(w_obj), space.w_unicode)
 
     def _join_check_item(self, space, w_obj):
-        if (space.isinstance_w(w_obj, space.w_str) or
+        if (space.isinstance_w(w_obj, space.w_bytes) or
             space.isinstance_w(w_obj, space.w_unicode)):
             return 0
         return 1
@@ -424,11 +432,11 @@ def plain_str2unicode(space, s):
                 raise OperationError(
                     space.w_UnicodeDecodeError,
                     space.newtuple([
-                    space.wrap('ascii'),
-                    space.wrap(s),
-                    space.wrap(i),
-                    space.wrap(i+1),
-                    space.wrap("ordinal not in range(128)")]))
+                    space.newtext('ascii'),
+                    space.newbytes(s),
+                    space.newint(i),
+                    space.newint(i+1),
+                    space.newtext("ordinal not in range(128)")]))
         assert False, "unreachable"
 
 
@@ -442,8 +450,8 @@ def getdefaultencoding(space):
 
 
 def _get_encoding_and_errors(space, w_encoding, w_errors):
-    encoding = None if w_encoding is None else space.str_w(w_encoding)
-    errors = None if w_errors is None else space.str_w(w_errors)
+    encoding = None if w_encoding is None else space.text_w(w_encoding)
+    errors = None if w_errors is None else space.text_w(w_errors)
     return encoding, errors
 
 
@@ -454,35 +462,26 @@ def encode_object(space, w_object, encoding, errors):
         w_encoder = space.sys.get_w_default_encoder()
     else:
         if errors is None or errors == 'strict':
-            try:
-                if encoding == 'ascii':
-                    u = space.unicode_w(w_object)
-                    eh = unicodehelper.raise_unicode_exception_encode
-                    return space.wrap(unicode_encode_ascii(
-                            u, len(u), None, errorhandler=eh))
-                if encoding == 'utf-8':
-                    u = space.unicode_w(w_object)
-                    eh = unicodehelper.raise_unicode_exception_encode
-                    return space.wrap(unicode_encode_utf_8(
-                            u, len(u), None, errorhandler=eh,
-                            allow_surrogates=True))
-            except unicodehelper.RUnicodeEncodeError as ue:
-                raise OperationError(space.w_UnicodeEncodeError,
-                                     space.newtuple([
-                    space.wrap(ue.encoding),
-                    space.wrap(ue.object),
-                    space.wrap(ue.start),
-                    space.wrap(ue.end),
-                    space.wrap(ue.reason)]))
+            if encoding == 'ascii':
+                u = space.unicode_w(w_object)
+                eh = unicodehelper.encode_error_handler(space)
+                return space.newbytes(unicode_encode_ascii(
+                        u, len(u), None, errorhandler=eh))
+            if encoding == 'utf-8':
+                u = space.unicode_w(w_object)
+                eh = unicodehelper.encode_error_handler(space)
+                return space.newbytes(unicode_encode_utf_8(
+                        u, len(u), None, errorhandler=eh,
+                        allow_surrogates=True))
         from pypy.module._codecs.interp_codecs import lookup_codec
-        w_encoder = space.getitem(lookup_codec(space, encoding), space.wrap(0))
+        w_encoder = space.getitem(lookup_codec(space, encoding), space.newint(0))
     if errors is None:
-        w_errors = space.wrap('strict')
+        w_errors = space.newtext('strict')
     else:
-        w_errors = space.wrap(errors)
+        w_errors = space.newtext(errors)
     w_restuple = space.call_function(w_encoder, w_object, w_errors)
-    w_retval = space.getitem(w_restuple, space.wrap(0))
-    if not space.isinstance_w(w_retval, space.w_str):
+    w_retval = space.getitem(w_restuple, space.newint(0))
+    if not space.isinstance_w(w_retval, space.w_bytes):
         raise oefmt(space.w_TypeError,
                     "encoder did not return an string object (type '%T')",
                     w_retval)
@@ -502,20 +501,20 @@ def decode_object(space, w_obj, encoding, errors):
                 eh = unicodehelper.decode_error_handler(space)
                 u = str_decode_ascii(     # try again, to get the error right
                     s, len(s), None, final=True, errorhandler=eh)[0]
-            return space.wrap(u)
+            return space.newunicode(u)
         if encoding == 'utf-8':
             s = space.charbuf_w(w_obj)
             eh = unicodehelper.decode_error_handler(space)
-            return space.wrap(str_decode_utf_8(
+            return space.newunicode(str_decode_utf_8(
                     s, len(s), None, final=True, errorhandler=eh,
                     allow_surrogates=True)[0])
     w_codecs = space.getbuiltinmodule("_codecs")
-    w_decode = space.getattr(w_codecs, space.wrap("decode"))
+    w_decode = space.getattr(w_codecs, space.newtext("decode"))
     if errors is None:
-        w_retval = space.call_function(w_decode, w_obj, space.wrap(encoding))
+        w_retval = space.call_function(w_decode, w_obj, space.newtext(encoding))
     else:
-        w_retval = space.call_function(w_decode, w_obj, space.wrap(encoding),
-                                       space.wrap(errors))
+        w_retval = space.call_function(w_decode, w_obj, space.newtext(encoding),
+                                       space.newtext(errors))
     return w_retval
 
 
@@ -537,7 +536,7 @@ def unicode_from_encoded_object(space, w_obj, encoding, errors):
 def unicode_from_object(space, w_obj):
     if space.is_w(space.type(w_obj), space.w_unicode):
         return w_obj
-    elif space.is_w(space.type(w_obj), space.w_str):
+    elif space.is_w(space.type(w_obj), space.w_bytes):
         w_res = w_obj
     else:
         w_unicode_method = space.lookup(w_obj, "__unicode__")
@@ -545,7 +544,7 @@ def unicode_from_object(space, w_obj):
         # test_unicode_conversion_with__str__
         if w_unicode_method is None:
             if space.isinstance_w(w_obj, space.w_unicode):
-                return space.wrap(space.unicode_w(w_obj))
+                return space.newunicode(space.unicode_w(w_obj))
             w_unicode_method = space.lookup(w_obj, "__str__")
         if w_unicode_method is not None:
             w_res = space.get_and_call_function(w_unicode_method, w_obj)
@@ -556,17 +555,17 @@ def unicode_from_object(space, w_obj):
     return unicode_from_encoded_object(space, w_res, None, "strict")
 
 
-def unicode_from_string(space, w_str):
+def unicode_from_string(space, w_bytes):
     # this is a performance and bootstrapping hack
     encoding = getdefaultencoding(space)
     if encoding != 'ascii':
-        return unicode_from_encoded_object(space, w_str, encoding, "strict")
-    s = space.str_w(w_str)
+        return unicode_from_encoded_object(space, w_bytes, encoding, "strict")
+    s = space.bytes_w(w_bytes)
     try:
         return W_UnicodeObject(s.decode("ascii"))
     except UnicodeDecodeError:
         # raising UnicodeDecodeError is messy, "please crash for me"
-        return unicode_from_encoded_object(space, w_str, "ascii", "strict")
+        return unicode_from_encoded_object(space, w_bytes, "ascii", "strict")
 
 
 class UnicodeDocstrings:
@@ -630,6 +629,9 @@ class UnicodeDocstrings:
     def __mod__():
         """x.__mod__(y) <==> x%y"""
 
+    def __rmod__():
+        """x.__rmod__(y) <==> y%x"""
+
     def __mul__():
         """x.__mul__(n) <==> x*n"""
 
@@ -1093,6 +1095,8 @@ W_UnicodeObject.typedef = TypeDef(
                             doc=UnicodeDocstrings.__format__.__doc__),
     __mod__ = interp2app(W_UnicodeObject.descr_mod,
                          doc=UnicodeDocstrings.__mod__.__doc__),
+    __rmod__ = interp2app(W_UnicodeObject.descr_rmod,
+                         doc=UnicodeDocstrings.__rmod__.__doc__),
     __getnewargs__ = interp2app(W_UnicodeObject.descr_getnewargs,
                                 doc=UnicodeDocstrings.__getnewargs__.__doc__),
     _formatter_parser = interp2app(W_UnicodeObject.descr_formatter_parser),
@@ -1130,10 +1134,10 @@ def unicode_to_decimal_w(space, w_unistr):
             if 0 < uchr < 256:
                 result[i] = chr(uchr)
             else:
-                w_encoding = space.wrap('decimal')
-                w_start = space.wrap(i)
-                w_end = space.wrap(i+1)
-                w_reason = space.wrap('invalid decimal Unicode string')
+                w_encoding = space.newtext('decimal')
+                w_start = space.newint(i)
+                w_end = space.newint(i+1)
+                w_reason = space.newtext('invalid decimal Unicode string')
                 raise OperationError(space.w_UnicodeEncodeError,
                                      space.newtuple([w_encoding, w_unistr,
                                                      w_start, w_end,
