@@ -44,11 +44,10 @@ class OptIntBounds(Optimization):
        redundant guards"""
 
     def propagate_forward(self, op):
-        dispatch_opt(self, op)
+        return dispatch_opt(self, op)
 
-    def opt_default(self, op):
-        assert not op.is_ovf()
-        self.emit_operation(op)
+    def propagate_postprocess(self, op):
+        return dispatch_postprocess(self, op)
 
     def propagate_bounds_backward(self, box):
         # FIXME: This takes care of the instruction where box is the reuslt
@@ -62,7 +61,9 @@ class OptIntBounds(Optimization):
             dispatch_bounds_ops(self, box)
 
     def _optimize_guard_true_false_value(self, op):
-        self.emit_operation(op)
+        return self.emit(op)
+
+    def _postprocess_guard_true_false_value(self, op):
         if op.getarg(0).type == 'i':
             self.propagate_bounds_backward(op.getarg(0))
 
@@ -70,18 +71,26 @@ class OptIntBounds(Optimization):
     optimize_GUARD_FALSE = _optimize_guard_true_false_value
     optimize_GUARD_VALUE = _optimize_guard_true_false_value
 
+    postprocess_GUARD_TRUE = _postprocess_guard_true_false_value
+    postprocess_GUARD_FALSE = _postprocess_guard_true_false_value
+    postprocess_GUARD_VALUE = _postprocess_guard_true_false_value
+
     def optimize_INT_OR_or_XOR(self, op):
         v1 = self.get_box_replacement(op.getarg(0))
-        b1 = self.getintbound(v1)
         v2 = self.get_box_replacement(op.getarg(1))
-        b2 = self.getintbound(v2)
         if v1 is v2:
             if op.getopnum() == rop.INT_OR:
                 self.make_equal_to(op, v1)
             else:
                 self.make_constant_int(op, 0)
-            return
-        self.emit_operation(op)
+            return None
+        return self.emit(op)
+
+    def postprocess_INT_OR_or_XOR(self, op):
+        v1 = self.get_box_replacement(op.getarg(0))
+        b1 = self.getintbound(v1)
+        v2 = self.get_box_replacement(op.getarg(1))
+        b2 = self.getintbound(v2)
         if b1.known_ge(IntBound(0, 0)) and \
            b2.known_ge(IntBound(0, 0)):
             r = self.getintbound(op)
@@ -91,11 +100,15 @@ class OptIntBounds(Optimization):
     optimize_INT_OR = optimize_INT_OR_or_XOR
     optimize_INT_XOR = optimize_INT_OR_or_XOR
 
+    postprocess_INT_OR = postprocess_INT_OR_or_XOR
+    postprocess_INT_XOR = postprocess_INT_OR_or_XOR
+
     def optimize_INT_AND(self, op):
+        return self.emit(op)
+
+    def postprocess_INT_AND(self, op):
         b1 = self.getintbound(op.getarg(0))
         b2 = self.getintbound(op.getarg(1))
-        self.emit_operation(op)
-
         r = self.getintbound(op)
         pos1 = b1.known_ge(IntBound(0, 0))
         pos2 = b2.known_ge(IntBound(0, 0))
@@ -107,7 +120,9 @@ class OptIntBounds(Optimization):
             r.make_le(b2)
 
     def optimize_INT_SUB(self, op):
-        self.emit_operation(op)
+        return self.emit(op)
+
+    def postprocess_INT_SUB(self, op):
         b1 = self.getintbound(op.getarg(0))
         b2 = self.getintbound(op.getarg(1))
         b = b1.sub_bound(b2)
@@ -118,8 +133,7 @@ class OptIntBounds(Optimization):
         arg1 = self.get_box_replacement(op.getarg(0))
         arg2 = self.get_box_replacement(op.getarg(1))
         if self.is_raw_ptr(arg1) or self.is_raw_ptr(arg2):
-            self.emit_operation(op)
-            return
+            return self.emit(op)
         v1 = self.getintbound(arg1)
         v2 = self.getintbound(arg2)
 
@@ -153,7 +167,9 @@ class OptIntBounds(Optimization):
                         arg2 = ConstInt(sum)
                         op = self.replace_op_with(op, rop.INT_ADD, args=[arg1, arg2])
 
-        self.emit_operation(op)
+        return self.emit(op)
+
+    def postprocess_INT_ADD(self, op):
         b1 = self.getintbound(op.getarg(0))
         b2 = self.getintbound(op.getarg(1))
         r = self.getintbound(op)
@@ -162,38 +178,37 @@ class OptIntBounds(Optimization):
             r.intersect(b)
 
     def optimize_INT_MUL(self, op):
+        return self.emit(op)
+
+    def postprocess_INT_MUL(self, op):
         b1 = self.getintbound(op.getarg(0))
         b2 = self.getintbound(op.getarg(1))
-        self.emit_operation(op)
         r = self.getintbound(op)
         b = b1.mul_bound(b2)
         if b.bounded():
             r.intersect(b)
 
     def optimize_CALL_PURE_I(self, op):
+        return self.emit(op)
+
+    def postprocess_CALL_PURE_I(self, op):
         # dispatch based on 'oopspecindex' to a method that handles
         # specifically the given oopspec call.
         effectinfo = op.getdescr().get_extra_info()
         oopspecindex = effectinfo.oopspecindex
         if oopspecindex == EffectInfo.OS_INT_PY_DIV:
-            self.opt_call_INT_PY_DIV(op)
-            return
+            self.post_call_INT_PY_DIV(op)
         elif oopspecindex == EffectInfo.OS_INT_PY_MOD:
-            self.opt_call_INT_PY_MOD(op)
-            return
-        self.emit_operation(op)
+            self.post_call_INT_PY_MOD(op)
 
-    def opt_call_INT_PY_DIV(self, op):
+    def post_call_INT_PY_DIV(self, op):
         b1 = self.getintbound(op.getarg(1))
         b2 = self.getintbound(op.getarg(2))
-        self.emit_operation(op)
         r = self.getintbound(op)
         r.intersect(b1.py_div_bound(b2))
 
-    def opt_call_INT_PY_MOD(self, op):
-        b1 = self.getintbound(op.getarg(1))
+    def post_call_INT_PY_MOD(self, op):
         b2 = self.getintbound(op.getarg(2))
-        self.emit_operation(op)
         if b2.is_constant():
             val = b2.getint()
             r = self.getintbound(op)
@@ -205,11 +220,13 @@ class OptIntBounds(Optimization):
                 r.make_le(IntBound(0, 0))
 
     def optimize_INT_LSHIFT(self, op):
+        return self.emit(op)
+
+    def postprocess_INT_LSHIFT(self, op):
         arg0 = self.get_box_replacement(op.getarg(0))
         b1 = self.getintbound(arg0)
         arg1 = self.get_box_replacement(op.getarg(1))
         b2 = self.getintbound(arg1)
-        self.emit_operation(op)
         r = self.getintbound(op)
         b = b1.lshift_bound(b2)
         r.intersect(b)
@@ -228,10 +245,15 @@ class OptIntBounds(Optimization):
         if b.has_lower and b.has_upper and b.lower == b.upper:
             # constant result (likely 0, for rshifts that kill all bits)
             self.make_constant_int(op, b.lower)
-        else:
-            self.emit_operation(op)
-            r = self.getintbound(op)
-            r.intersect(b)
+            return None
+        return self.emit(op)
+
+    def postprocess_INT_RSHIFT(self, op):
+        b1 = self.getintbound(op.getarg(0))
+        b2 = self.getintbound(op.getarg(1))
+        b = b1.rshift_bound(b2)
+        r = self.getintbound(op)
+        r.intersect(b)
 
     def optimize_GUARD_NO_OVERFLOW(self, op):
         lastop = self.last_emitted_operation
@@ -257,7 +279,7 @@ class OptIntBounds(Optimization):
                 self.pure_from_args(rop.INT_SUB, [args[0], result], args[1])
             #elif opnum == rop.INT_MUL_OVF:
             #    self.pure(rop.INT_MUL, args[:], result)
-            self.emit_operation(op)
+            return self.emit(op)
 
     def optimize_GUARD_OVERFLOW(self, op):
         # If INT_xxx_OVF was replaced by INT_xxx, *but* we still see
@@ -270,7 +292,7 @@ class OptIntBounds(Optimization):
             raise InvalidLoop('An INT_xxx_OVF was proven not to overflow but' +
                               'guarded with GUARD_OVERFLOW')
 
-        self.emit_operation(op)
+        return self.emit(op)
 
     def optimize_INT_ADD_OVF(self, op):
         b1 = self.getintbound(op.getarg(0))
@@ -281,7 +303,12 @@ class OptIntBounds(Optimization):
             # by optimize_GUARD_NO_OVERFLOW; if we see instead an
             # optimize_GUARD_OVERFLOW, then InvalidLoop.
             op = self.replace_op_with(op, rop.INT_ADD)
-        self.emit_operation(op) # emit the op
+        return self.emit(op)
+
+    def postprocess_INT_ADD_OVF(self, op):
+        b1 = self.getintbound(op.getarg(0))
+        b2 = self.getintbound(op.getarg(1))
+        resbound = b1.add_bound(b2)
         r = self.getintbound(op)
         r.intersect(resbound)
 
@@ -292,11 +319,18 @@ class OptIntBounds(Optimization):
         b1 = self.getintbound(arg1)
         if arg0.same_box(arg1):
             self.make_constant_int(op, 0)
-            return
+            return None
         resbound = b0.sub_bound(b1)
         if resbound.bounded():
             op = self.replace_op_with(op, rop.INT_SUB)
-        self.emit_operation(op) # emit the op
+        return self.emit(op)
+
+    def postprocess_INT_SUB_OVF(self, op):
+        arg0 = self.get_box_replacement(op.getarg(0))
+        arg1 = self.get_box_replacement(op.getarg(1))
+        b0 = self.getintbound(arg0)
+        b1 = self.getintbound(arg1)
+        resbound = b0.sub_bound(b1)
         r = self.getintbound(op)
         r.intersect(resbound)
 
@@ -306,7 +340,12 @@ class OptIntBounds(Optimization):
         resbound = b1.mul_bound(b2)
         if resbound.bounded():
             op = self.replace_op_with(op, rop.INT_MUL)
-        self.emit_operation(op)
+        return self.emit(op)
+
+    def postprocess_INT_MUL_OVF(self, op):
+        b1 = self.getintbound(op.getarg(0))
+        b2 = self.getintbound(op.getarg(1))
+        resbound = b1.mul_bound(b2)
         r = self.getintbound(op)
         r.intersect(resbound)
 
@@ -320,7 +359,7 @@ class OptIntBounds(Optimization):
         elif b1.known_ge(b2) or arg1 is arg2:
             self.make_constant_int(op, 0)
         else:
-            self.emit_operation(op)
+            return self.emit(op)
 
     def optimize_INT_GT(self, op):
         arg1 = self.get_box_replacement(op.getarg(0))
@@ -332,7 +371,7 @@ class OptIntBounds(Optimization):
         elif b1.known_le(b2) or arg1 is arg2:
             self.make_constant_int(op, 0)
         else:
-            self.emit_operation(op)
+            return self.emit(op)
 
     def optimize_INT_LE(self, op):
         arg1 = self.get_box_replacement(op.getarg(0))
@@ -344,7 +383,7 @@ class OptIntBounds(Optimization):
         elif b1.known_gt(b2):
             self.make_constant_int(op, 0)
         else:
-            self.emit_operation(op)
+            return self.emit(op)
 
     def optimize_INT_GE(self, op):
         arg1 = self.get_box_replacement(op.getarg(0))
@@ -356,13 +395,13 @@ class OptIntBounds(Optimization):
         elif b1.known_lt(b2):
             self.make_constant_int(op, 0)
         else:
-            self.emit_operation(op)
+            return self.emit(op)
 
     def optimize_INT_EQ(self, op):
         arg0 = self.get_box_replacement(op.getarg(0))
+        b1 = self.getintbound(arg0)
         arg1 = self.get_box_replacement(op.getarg(1))
-        b1 = self.getintbound(op.getarg(0))
-        b2 = self.getintbound(op.getarg(1))
+        b2 = self.getintbound(arg1)
         if b1.known_gt(b2):
             self.make_constant_int(op, 0)
         elif b1.known_lt(b2):
@@ -370,7 +409,7 @@ class OptIntBounds(Optimization):
         elif arg0.same_box(arg1):
             self.make_constant_int(op, 1)
         else:
-            self.emit_operation(op)
+            return self.emit(op)
 
     def optimize_INT_NE(self, op):
         arg0 = self.get_box_replacement(op.getarg(0))
@@ -384,14 +423,14 @@ class OptIntBounds(Optimization):
         elif arg0 is arg1:
             self.make_constant_int(op, 0)
         else:
-            self.emit_operation(op)
+            return self.emit(op)
 
     def optimize_INT_FORCE_GE_ZERO(self, op):
         b = self.getintbound(op.getarg(0))
         if b.known_ge(IntBound(0, 0)):
             self.make_equal_to(op, op.getarg(0))
         else:
-            self.emit_operation(op)
+            return self.emit(op)
 
     def optimize_INT_SIGNEXT(self, op):
         b = self.getintbound(op.getarg(0))
@@ -402,29 +441,43 @@ class OptIntBounds(Optimization):
         if bounds.contains_bound(b):
             self.make_equal_to(op, op.getarg(0))
         else:
-            self.emit_operation(op)
-            bres = self.getintbound(op)
-            bres.intersect(bounds)
+            return self.emit(op)
+
+    def postprocess_INT_SIGNEXT(self, op):
+        numbits = op.getarg(1).getint() * 8
+        start = -(1 << (numbits - 1))
+        stop = 1 << (numbits - 1)
+        bounds = IntBound(start, stop - 1)
+        bres = self.getintbound(op)
+        bres.intersect(bounds)
 
     def optimize_ARRAYLEN_GC(self, op):
+        return self.emit(op)
+
+    def postprocess_ARRAYLEN_GC(self, op):
         array = self.ensure_ptr_info_arg0(op)
-        self.emit_operation(op)
         self.optimizer.setintbound(op, array.getlenbound(None))
 
     def optimize_STRLEN(self, op):
-        self.emit_operation(op)
+        return self.emit(op)
+
+    def postprocess_STRLEN(self, op):
         self.make_nonnull_str(op.getarg(0), vstring.mode_string)
         array = self.getptrinfo(op.getarg(0))
         self.optimizer.setintbound(op, array.getlenbound(vstring.mode_string))
 
     def optimize_UNICODELEN(self, op):
-        self.emit_operation(op)
+        return self.emit(op)
+
+    def postprocess_UNICODELEN(self, op):
         self.make_nonnull_str(op.getarg(0), vstring.mode_unicode)
         array = self.getptrinfo(op.getarg(0))
         self.optimizer.setintbound(op, array.getlenbound(vstring.mode_unicode))
 
     def optimize_STRGETITEM(self, op):
-        self.emit_operation(op)
+        return self.emit(op)
+
+    def postprocess_STRGETITEM(self, op):
         v1 = self.getintbound(op)
         v2 = self.getptrinfo(op.getarg(0))
         intbound = self.getintbound(op.getarg(1))
@@ -436,7 +489,9 @@ class OptIntBounds(Optimization):
         v1.make_lt(IntUpperBound(256))
 
     def optimize_GETFIELD_RAW_I(self, op):
-        self.emit_operation(op)
+        return self.emit(op)
+
+    def postprocess_GETFIELD_RAW_I(self, op):
         descr = op.getdescr()
         if descr.is_integer_bounded():
             b1 = self.getintbound(op)
@@ -449,12 +504,24 @@ class OptIntBounds(Optimization):
     optimize_GETFIELD_GC_R = optimize_GETFIELD_RAW_I
     optimize_GETFIELD_GC_F = optimize_GETFIELD_RAW_I
 
+    postprocess_GETFIELD_RAW_F = postprocess_GETFIELD_RAW_I
+    postprocess_GETFIELD_RAW_R = postprocess_GETFIELD_RAW_I
+    postprocess_GETFIELD_GC_I = postprocess_GETFIELD_RAW_I
+    postprocess_GETFIELD_GC_R = postprocess_GETFIELD_RAW_I
+    postprocess_GETFIELD_GC_F = postprocess_GETFIELD_RAW_I
+
     optimize_GETINTERIORFIELD_GC_I = optimize_GETFIELD_RAW_I
     optimize_GETINTERIORFIELD_GC_R = optimize_GETFIELD_RAW_I
     optimize_GETINTERIORFIELD_GC_F = optimize_GETFIELD_RAW_I
 
+    postprocess_GETINTERIORFIELD_GC_I = postprocess_GETFIELD_RAW_I
+    postprocess_GETINTERIORFIELD_GC_R = postprocess_GETFIELD_RAW_I
+    postprocess_GETINTERIORFIELD_GC_F = postprocess_GETFIELD_RAW_I
+
     def optimize_GETARRAYITEM_RAW_I(self, op):
-        self.emit_operation(op)
+        return self.emit(op)
+
+    def postprocess_GETARRAYITEM_RAW_I(self, op):
         descr = op.getdescr()
         if descr and descr.is_item_integer_bounded():
             intbound = self.getintbound(op)
@@ -466,8 +533,15 @@ class OptIntBounds(Optimization):
     optimize_GETARRAYITEM_GC_F = optimize_GETARRAYITEM_RAW_I
     optimize_GETARRAYITEM_GC_R = optimize_GETARRAYITEM_RAW_I
 
+    postprocess_GETARRAYITEM_RAW_F = postprocess_GETARRAYITEM_RAW_I
+    postprocess_GETARRAYITEM_GC_I = postprocess_GETARRAYITEM_RAW_I
+    postprocess_GETARRAYITEM_GC_F = postprocess_GETARRAYITEM_RAW_I
+    postprocess_GETARRAYITEM_GC_R = postprocess_GETARRAYITEM_RAW_I
+
     def optimize_UNICODEGETITEM(self, op):
-        self.emit_operation(op)
+        return self.emit(op)
+
+    def postprocess_UNICODEGETITEM(self, op):
         b1 = self.getintbound(op)
         b1.make_ge(IntLowerBound(0))
         v2 = self.getptrinfo(op.getarg(0))
@@ -632,5 +706,6 @@ class OptIntBounds(Optimization):
 
 
 dispatch_opt = make_dispatcher_method(OptIntBounds, 'optimize_',
-        default=OptIntBounds.opt_default)
+                                      default=OptIntBounds.emit)
 dispatch_bounds_ops = make_dispatcher_method(OptIntBounds, 'propagate_bounds_')
+dispatch_postprocess = make_dispatcher_method(OptIntBounds, 'postprocess_')
